﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;

using System.Resources;
using System.Threading;
using System.Reflection;
using System.Diagnostics;
using System.IO;
using System.Globalization;

using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;

/////////////////////////////////////////////////////////////////////////////////
/// <summary>
/// R付き連続線分作成コマンド
/// 概要
/// 　連続する線分とそれらをつなぐ各頂点部分を円弧(R)にした図形を作図する RootPro CAD のアドインプログラムです。
/// ※使用許諾
/// ・このアドインを実行した結果についての一切の責任はお使いになられる方にあります。
/// ・このアドインは、プログラムソース自体を改変・流用することが可能です。
/// 　また、このプログラムソースを元に独自のアドインやソフトウェアを作成し、無償・有償を問わず第三者に配布することも認めます。
/// 　ただし、改変・流用する場合は、株式会社ルートプロ(RootPro Corporation)作成の
/// 　当プログラムソースを参考にしていることを第三者にわかるように明記してください。
///
/// 　Copyright @ 2023 RootPro Co.,Ltd. (Japan)
/// </summary>
/// <summary locale = "en">
/// Continuous line segment creation command with R
/// overview
/// This is a RootPro CAD add-in program that draws continuous line segments and shapes with arcs (R) at each vertex that connects them.
///
/// License
///  The user is solely responsible for the result of executing this add-in.
///  This add-in can modify or divert the program source itself.
///  We also allow you to create your own add-ins and software based on this program source and distribute them to third parties regardless of whether they are free or paid.
/// 
///  Copyright @ 2023 RootPro Co., Ltd. (Japan)
/// </summary>


namespace RCAddInRPolylineCS
{

    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
        // 設定値
        AddInSettings addInSettings = null;

        string RPolylineCommandName = "RootPro.AddIn.RPolylineCommand";     //コマンド名 

        // コマンドパラメータID
        const int radiusParameterId = 1;      // Rの半径パラメータID
        const int pointListParameterId = 2;   // 点列パラメータID

         // エラー番号
        const int RC_ERR_SUCCESS = 1; // エラーなし（成功）
        const int RC_ERR_NONPOINTS = -1; // エラーなし（点列なし）

        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            // ファイルから設定値を読み込む
            if(!LoadAddInSettings(out addInSettings))
                addInSettings = new AddInSettings();// 読み込めなかった場合は初期値

            // 多言語対応リソースのDLLファイルをロード
            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

			CommandManager.AddMacroCommand("RCAddInRPolylineCS", Properties.Resources.CommandName, MacroCommand);// リソースファイルを追加してコマンド名などを定義しておく

            // 図形コマンドを追加する
            Command command = CommandManager.AddCommand(RPolylineCommandName, Properties.Resources.CommandName);// AddIn のコマンド名をセット
	        AddParameters(command);// コマンドパラメーターセット

            // 開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            // 設定値をファイルに書き込む
            Debug.Assert(addInSettings != null);
            SaveAddInSettings(addInSettings);

            // コマンドの削除
            CommandManager.RemoveMacroCommand(MacroCommand);
            CommandManager.RemoveCommand(RPolylineCommandName);
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        /// <summary>
        /// 外部フォルダのアセンブリを動的にロード
        /// 多言語対応リソースのDLLファイルをロード
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
		{
			AssemblyName assemblyName = new AssemblyName(args.Name);

			foreach (Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
			{
				if (AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
					return assembly;
			}

			if (assemblyName.Name.EndsWith(".resources"))
			{
				if (assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
				{
					string dllPath = Path.GetDirectoryName(this.AddInFullPath);
					dllPath = Path.Combine(dllPath, assemblyName.CultureName);
					dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

					if (File.Exists(dllPath))
					{
						try
						{
							return Assembly.LoadFrom(dllPath);
						}
						catch
						{
						}
					}
				}
			}

			return null;
		}


		 private void MacroCommand()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(RPolylineCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(command_ParameterChanged);

            //マウス移動イベントの追加
            command.MouseMove += new CommandMouseMoveEventHandler(Command_MouseMove);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(command_ParameterChanged);

            //マウス移動イベントの削除
            command.MouseMove -= new CommandMouseMoveEventHandler(Command_MouseMove);
        }

        /// <summary>
        /// パラメータを追加する
        /// </summary>
        /// <param name="command"></param>
        /// <param name="InputParameterType"></param>
        private void AddParameters(Command command)
        {
            // コマンドウィンドウの各行にパラメータを追加
            // メモ：コマンドウィンドウでコマンドのパラメーターを用意する
            // 各パラメーターは ParameterItemType がありそれぞれ扱いたい値によって使い分ける
            // パラメータの種類
            // ParameterItemType.Point      --- PointParameterItem;      // 座標パラメータ         -- 座標値を入力できます。マウスで図面上から指定することも可能です。
            // ParameterItemType.PointList  --- PointListParameterItem;  // 点列パラメータ         -- 連続した複数の座標値を入力できます。
            // ParameterItemType.Length     --- LengthParameterItem;     // 長さパラメータ         -- 長さを入力できます。
            // ParameterItemType.Angle      --- AngleParameterItem;      // 角度パラメータ         -- 角度を入力できます。
            // ParameterItemType.Double     --- DoubleParameterItem;     // 倍精度浮動小数点数パラメータ
            // ParameterItemType.Integer    --- IntegerParameterItem;    // 整数パラメータ
            // ParameterItemType.Scale      --- ScaleParameterItem;      // 倍率パラメータ
            // ParameterItemType.Text       --- TextParameterItem;       // 文字パラメータ         -- 文字列を入力できます。
            // ParameterItemType.List       --- ListParameterItem;       // リストパラメータ       -- 選択肢をセットして選ぶことができます。
            // ParameterItemType.Shape      --- ShapeParameterItem;      // 図形パラメータ         -- 図形が選択できます。
            // ParameterItemType.ShapeGroup --- ShapeGroupParameterItem; // 図形グループパラメータ -- 複数の図形が選択できます。
            // ParameterItemType.Multiple   --- MultipleParameterItem;   // 複合パラメータ


            // パラメータを追加する
            // R半径
            command.ParameterGroups.Add(ParameterItemType.Length, radiusParameterId, Properties.Resources.RadiusParameterName);
            LengthParameterItem radiusParameterItem = (LengthParameterItem)command.GetParameterItemByID(radiusParameterId); // パラメータアイテム
            radiusParameterItem.Description = Properties.Resources.RadiusParameterDescription;
            radiusParameterItem.ValueSaveMode = ParameterValueSaveMode.Always;
            radiusParameterItem.Length = addInSettings.radiusParameterValue;// 設定の値を初期値としてセット

            // 「点列」パラメーター 
            command.ParameterGroups.Add(ParameterItemType.PointList, pointListParameterId, Properties.Resources.PointListParameterName);
            PointListParameterItem pointListParameterItem = (PointListParameterItem)command.GetParameterItemByID(pointListParameterId);
            pointListParameterItem.Description = Properties.Resources.PointListParameterDescription;
        }

        /// <summary>
        /// パラメータの変化のイベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            Command command = (Command)sender;

            // 半径パラメータの取得
            LengthParameterItem radiusParameterItem = (LengthParameterItem)command.GetParameterItemByID(radiusParameterId);
            double radius = 0.0;
            if(radiusParameterItem != null)
                radius = radiusParameterItem.Length;

            if(radius != 0.0 && radius <= Precision.Confusion)
                return;

            // 点列パラメーターの取得
            PointListParameterItem pointListParameterItem = (PointListParameterItem)command.GetParameterItemByID(pointListParameterId);
            if(pointListParameterItem == null)
                return;

            // 点列の入力が完了していない場合
            if(pointListParameterItem.IsEmpty || !pointListParameterItem.Complete)
                return;

            Point2d[] pointList = pointListParameterItem.Points;
            if(pointList.Length <= 1)
                return;

            // 図形作成
            List<Shape> shapes = null;
            int stat = CreateCommandShapes(pointList, radius, out shapes);
            if(stat < 0)
            {
                //ShowErrMessage(stat);

                e.CommandState = CommandState.ParameterError;
                return;
            }

            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;

            // UNDO 可能な操作を開始
            doc.UndoManager.BeginUndoUnit();

            // 図面に図形追加
            foreach(Shape shape in shapes)
            {
                drawing.Shapes.Add(shape);
            }

            // UNDO 可能な操作を終了
            doc.UndoManager.EndUndoUnit();

            // コマンド実行
            e.CommandState = CommandState.Execute;

            // 「R半径」パラメータの値を設定に保持
            addInSettings.radiusParameterValue = radiusParameterItem.Length;
        }

        /// <summary>
        /// マウス移動イベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Command_MouseMove(Object sender, CommandMouseMoveEventArgs e)
        {
            Command command = (Command)sender;
            // Command.GetTemporaryParameterItemByID を使用してコマンドの一時的なパラメータアイテムを取得します。
            // 一時的なパラメータアイテムからは、すでにパラメータの入力が確定している場合は確定している入力値が、
            // 入力が確定していない場合はマウスの現在位置などの一時的な値が取得できます。

            // 半径パラメータの取得
            double radius = 0.0;
            LengthParameterItem radiusParameterItem = (LengthParameterItem)command.GetTemporaryParameterItemByID(radiusParameterId);
            if(radiusParameterItem == null)
                radius = 0.0;
            else
                radius = radiusParameterItem.Length;

            if(radius != 0.0 && radius <= Precision.Confusion)
                return;

            // 点列パラメーターの取得
            PointListParameterItem pointListParameterItem = (PointListParameterItem)command.GetTemporaryParameterItemByID(pointListParameterId);
            if(pointListParameterItem == null)
                return;

            Point2d[] pointList = pointListParameterItem.Points;
            if(pointList.Length <= 1)
                return;

             // 図形作成
            List<Shape> shapes = null;
            int stat = CreateCommandShapes(pointList, radius, out shapes);
            if(stat != RC_ERR_SUCCESS)
                return;

            // ラバーバンド図形リストに追加
            List<Shape> rubberBandShapes = new List<Shape>();
            for(int i = 0; i < shapes.Count; i++)
                rubberBandShapes.Add(shapes[i]);

            // ラバーバンド図形描画
            if(rubberBandShapes.Count > 0)
                command.DrawShapeRubberBand(rubberBandShapes.ToArray());
        }

        /// <summary>
        /// 出力図形の作成
        /// </summary>
        /// <param name="pointList"></param>
        /// <param name="radius"></param>
        /// <param name="shapes"></param>
        /// <returns></returns>
        private int CreateCommandShapes(Point2d[] pointList, double radius, out List<Shape> shapes)
        {
            shapes = null;

            if(pointList.Length < 2)
                return RC_ERR_NONPOINTS;

            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;
            // 書込みレイヤ
            Layer currentLayer = doc.LayerTable.CurrentLayer;

            // 点列から連続線分のリストを作成
            List<Shape> baseLineShapes = new List<Shape>();
            for(int i = 1; i < pointList.Length; i++)
            {
                Point2d sp = pointList[i - 1];
                Point2d ep = pointList[i];
                LineShape lineShape = ShapeFactory.CreateLine(sp, ep);
                doc.BasicShapeSettings.CopyToShape(lineShape);// 既定値をセット
                lineShape.Layer = currentLayer;// レイヤ
                baseLineShapes.Add(lineShape);
            }

            // 2つずつの線分フィレットをして、結果の図形をリストに追加していく
            List<Shape> filletedShapes = new List<Shape>();
            if(radius > Precision.Confusion && baseLineShapes.Count > 1)
            {
                int forLoopStartIndex = 1; // for 文のスタートインデックスを1
                LineShape lastTrimmedShape2 = (LineShape)baseLineShapes[0];// 前回トリムされた2番目の図形を最初の線分にしておく
                // 最初の点と最後の点が同じ場合
                if(pointList.Length > 3)
                {
                    Point2d firstPoint = pointList[0];
                    Point2d lastPoint = pointList[pointList.Length - 1];
                    if(Geometry.IsPointEquals(firstPoint, lastPoint))
                    {
                        forLoopStartIndex = 0;// for 文のスタートインデックスを0
                        lastTrimmedShape2 = (LineShape)baseLineShapes[baseLineShapes.Count - 1];// 前回トリムされた2番目の図形を最後の線分にしておく
                    }
                }

                for(int i = forLoopStartIndex; i < baseLineShapes.Count; i++)
                {
                    LineShape shape1 = lastTrimmedShape2;
                    LineShape shape2 = (LineShape)baseLineShapes[i];
                    if(shape1 == null)
                    {
                        lastTrimmedShape2 = shape2;
                        continue;
                    }

                    // ピック点を中点とする
                    Point2d pkPoint1 = new Point2d((shape1.StartPoint.X + shape1.EndPoint.X) * 0.5, (shape1.StartPoint.Y + shape1.EndPoint.Y) * 0.5);
                    Point2d pkPoint2 = new Point2d((shape2.StartPoint.X + shape2.EndPoint.X) * 0.5, (shape2.StartPoint.Y + shape2.EndPoint.Y) * 0.5);

                    // フィレットを実行
                    ShapeFillet.TrimMode = TrimMode.Both;
                    ShapeFillet.Radius = radius;
                    FilletResult filletResult = ShapeFillet.Execute(shape1, pkPoint1, shape2, pkPoint2);
                    if(filletResult.Succeeded)
                    {
                        LineShape trimmedShape1 = null;
                        ArcShape filletShape = null;
                        LineShape trimmedShape2 = null;

                        if(filletResult.ShapeTrimState1 != ShapeTrimState.Deleted)
                            trimmedShape1 = (LineShape)filletResult.TrimmedShape1;

                        if(filletResult.ShapeTrimState2 != ShapeTrimState.Deleted)
                            trimmedShape2 = (LineShape)filletResult.TrimmedShape2;

                        if(trimmedShape1 != null)
                        {
                            // trimmedShape1 が最初のループで、最後の線分としてトリムされたものの場合
                            if(i == 0)
                            {
                                baseLineShapes.Remove(shape1);// 最後の線分を削除して
                                baseLineShapes.Add(trimmedShape1);// トリムされた線分を最後に追加しておく
                            }
                            else
                            {
                                filletedShapes.Add(trimmedShape1);
                            }
                        }

                        filletShape = filletResult.FilletShape;
                        if(filletShape != null)
                        {
                            doc.BasicShapeSettings.CopyToShape(filletShape);
                            filletShape.Layer = currentLayer;// レイヤ
                            filletedShapes.Add(filletShape);
                        }
                        
                        // 最後の場合
                        if(i == baseLineShapes.Count - 1 && trimmedShape2 != null)
                            filletedShapes.Add(trimmedShape2);

                        lastTrimmedShape2 = trimmedShape2;
                    }
                }

            }
            else
            {
                filletedShapes.AddRange(baseLineShapes);
            }

            // フィレットされた結果の図形
            shapes = new List<Shape>();// 作成図形
            shapes.AddRange(filletedShapes);
            
            return RC_ERR_SUCCESS;
        }

    }
}
